#!/usr/bin/env python3

# READ IN BTF / BRDF IMAGES FROM DATASET AND WRITE OUT TEXEL "IMAGES" (basically just ABRDFs saved as EXRs) CONTAINING THE LIST OF REFLECTANCE VALUES WITH ASSOCIATED ANGLES FOR ONE POINT ON THE MATERIAL
# OUTPUT HAS 7 CHANNELS : R, G, B, VIew azimuth, view elevation, light azimuth, light elevation 

# If there is an error in the EXR reading function, replace float16 by float32 (Matlab writes EXRs in half)

#Example of user parameters
datadir = "/home/gilles/Desktop/phd/datasets/deeplearning/data-original/bonnbtftextures/felt12"
outdir = "/home/gilles/Desktop/phd/datasets/deeplearning/data-preprocessed/bonnbtftexels/felt12"
pxIStart = 0   
pxJStart = 0
pxILen = 100
pxJLen = 100
outindex0 = 0



import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np 
import sys
import array
import OpenEXR
import Imath
import glob
import os 
from functools import reduce
from random import shuffle  
from skimage.transform import resize
def load_exr(pth):
    imexr = OpenEXR.InputFile(pth)
    (r, g, b) = imexr.channels("RGB")
    dw = imexr.header()['dataWindow']
    imsize = (dw.max.x - dw.min.x + 1, dw.max.y - dw.min.y + 1)
    im = np.zeros( (imsize[0],imsize[1], 3), dtype=np.float32) 
    
    rnp = np.reshape( np.frombuffer(r, dtype=np.float16), imsize)
    gnp = np.reshape( np.frombuffer(g, dtype=np.float16), imsize)
    bnp = np.reshape( np.frombuffer(b, dtype=np.float16), imsize)
    
    im[:,:,0] = rnp.astype(np.float32)
    im[:,:,1] = gnp.astype(np.float32)
    im[:,:,2] = bnp.astype(np.float32) 
    #print(imexr.header()['camera azimuth'], imexr.header()['camera elevation'], imexr.header()['light azimuth'], imexr.header()['light elevation'])
    return im

def textures2texels(dsdir, outdir, _pxIStart, _pxJStart, _pxILen, _pxJLen, outindex0=0, multiplier=1.0):
    if not os.path.exists(outdir):
        os.makedirs(outdir) 
    flist = sorted(glob.glob(dsdir+"/*.exr"))
    numims = len(flist)
    print( dsdir )
    print( str(numims) +" images found") 
    # read all images in dataset
    im = load_exr(flist[0]) 
    print(im.shape)
    angles = []
    dataset = np.zeros( (numims, im.shape[0],im.shape[1], 3), dtype=np.float32)   
    for i,f in enumerate(flist): 
        fn = f.split("-")
#         print(fn)
        angles.append( [fn[2], fn[3], fn[4], fn[5]])  
#         print( [fn[2], fn[3], fn[4], fn[5]] )
        dataset[i, :,:,:] = resize(load_exr(f), (im.shape[0], im.shape[1], 3), preserve_range=True)  
    print("read in dataset ", dataset.shape)
    
    dataset = np.swapaxes( dataset, 1, 2) # BECAUSE MATLAB DOES THE RESHAPING HEIGHT FIRST
    print("dataset after transposing ", dataset.shape)
    
    outset = np.zeros((_pxILen*_pxJLen, numims,1,7), dtype=np.float32) # r,g,b,vaz,vel,laz,lel
    
    # for each pixel, write EXR image containing all BTF values & associated angles (=ABRDF)
    for i in range(numims):  
        outset[:,i,0,0:3] = np.reshape(dataset[i, _pxIStart:_pxIStart+_pxILen, _pxJStart:_pxJStart+_pxJLen, :], (_pxILen*_pxJLen,3))
        outset[:,i,0,3:7] = np.tile(np.array([np.float32(angles[i][0]), np.float32(angles[i][1]), np.float32(angles[i][2]), np.float32(angles[i][3])]), (_pxILen*_pxJLen, 1))  
            
    print(outset.shape)
    del dataset 
    print("texels reshaped. writing output")
    for i in range(_pxILen*_pxJLen):
        im = outset[i, :,:,:]      
        im = np.clip(im, 0, None) 
        if i%1000==0:
            print(i)
            
#        print("image "+str(i), im.min(), im.mean(), im.max(), im.shape)  
        outname = outdir + "/px-" + str(i + outindex0).zfill(6) + ".exr"  
        outhead = OpenEXR.Header(im.shape[0], im.shape[1]) 
        outhead['channels'] = {'R': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)), 'B': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)), 
                               'G': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)), 'C': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)), 
                               'D': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)), 'E': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)), 
                               'F': Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)) }
        (r, g, b, vaz, vel, laz, lel) = [ im[:,:,c].tostring() for c in range(7)]   
        #print(outhead) 
        out = OpenEXR.OutputFile(outname, outhead)
        out.writePixels({'R' : r, 'G' : g, 'B' : b, 'C' : vaz, 'D' : vel, 'E' : laz, 'F' : lel})
        
textures2texels(datadir, outdir, pxIStart, pxJStart, pxILen, pxJLen, outindex0)  
